@@ -1,13 +1,24 @@
-/* Copyright (C) 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevcgml.c */
@@ -670,8 +681,9 @@ cgm_CHARACTER_SET_INDEX(cgm_state *st, cgm_int index)
 	END;
 }
 
+/* See gdevcgml.c for why this isn't named cgm_ALTERNATE_.... */
 cgm_result
-cgm_ALTERNATE_CHARACTER_SET_INDEX(cgm_state *st, cgm_int index)
+cgm_ALT_CHARACTER_SET_INDEX(cgm_state *st, cgm_int index)
 {	OP(ALTERNATE_CHARACTER_SET_INDEX);
 	IX(index);  st->alternate_character_set_index = index;
 	END;
@@ -956,13 +968,13 @@ put_string(cgm_state *st, const char *data, uint length)
 	{	put_byte(st, 255);
 		while ( length > 32767 )
 		{	put_int(st, 65535, 2);
-			put_bytes(st, data, 32767);
+			put_bytes(st, (const byte *)data, 32767);
 			data += 32767;
 			length -= 32767;
 		}
 	}
 	put_byte(st, length);
-	put_bytes(st, data, length);
+	put_bytes(st, (const byte *)data, length);
 }
 
 /* Put a color. */
