@@ -1,13 +1,24 @@
-/* Copyright (C) 1989, 1992, 1993, 1994 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevpcfb.c */
@@ -42,7 +53,7 @@ private dev_proc_map_color_rgb(ega01_map_color_rgb);
 #   define ega_map_color_rgb ega2_map_color_rgb
 # endif
 #endif
-#define ega_std_procs\
+#define ega_std_procs(get_params, put_params)\
 	ega_open,\
 	NULL,			/* get_initial_matrix */\
 	NULL,			/* sync_output */\
@@ -55,29 +66,34 @@ private dev_proc_map_color_rgb(ega01_map_color_rgb);
 	ega_copy_mono,\
 	ega_copy_color,\
 	NULL,			/* draw_line */\
-	ega_get_bits
+	ega_get_bits,\
+	get_params,\
+	put_params,\
+	NULL,			/* map_cmyk_color */\
+	NULL,			/* get_xfont_procs */\
+	NULL,			/* get_xfont_device */\
+	NULL,			/* map_rgb_alpha_color */\
+	gx_page_device_get_page_device
 
 private gx_device_procs ega_procs = {
-	ega_std_procs
+	ega_std_procs(NULL, NULL)
 };
 
 private dev_proc_get_params(svga16_get_params);
 private dev_proc_put_params(svga16_put_params);
 private gx_device_procs svga16_procs = {
-	ega_std_procs,
-	svga16_get_params,
-	svga16_put_params
+	ega_std_procs(svga16_get_params, svga16_put_params)
 };
 
 /* All the known instances */
 		/* EGA */
-gx_device_ega gs_ega_device =
+gx_device_ega far_data gs_ega_device =
 	ega_device("ega", ega_procs, 80, 350, 48.0/35.0, 0x10);
 		/* VGA */
-gx_device_ega gs_vga_device =
+gx_device_ega far_data gs_vga_device =
 	ega_device("vga", ega_procs, 80, 480, 1.0, 0x12);
 		/* Generic SuperVGA, 800x600, 16-color mode */
-gx_device_ega gs_svga16_device =
+gx_device_ega far_data gs_svga16_device =
 	ega_device("svga16", svga16_procs, 100, 600, 1.0, 0x29 /*Tseng*/);
 
 /* Save the BIOS state */
@@ -744,32 +760,46 @@ tile:	yleft = tile_height - ymod;
 /* Read scan lines back from the frame buffer. */
 int
 ega_get_bits(gx_device *dev, int y, byte *data, byte **actual_data)
-{	uint bytes_per_row = dev->width >> 1;
-	int plane;
-	ushort *dest = (ushort *)data;
-	if ( y < 0 || y >= dev->height )
-		return gs_error_rangecheck;
-	memset(data, 0, bytes_per_row);
-	for ( plane = 0; plane < 4; plane++ )
-	   {	fb_ptr src = mk_fb_ptr(0, y);
-		int i;
-		/* Plane 0 is the least significant plane. */
-		/* We know we're on a little-endian machine.... */
-		static const ushort spread4[16] =
-		   {	0x0, 0x800, 0x8000, 0x8800,
-			0x8, 0x808, 0x8008, 0x8808,
-			0x80, 0x880, 0x8080, 0x8880,
-			0x88, 0x888, 0x8088, 0x8888
-		   };
-		set_g_read_plane(plane);
-		for ( i = 0; i < dev->width; i += 8, src++, dest += 2 )
-		   {	byte b = *src;
-			dest[0] = (dest[0] >> 1) + spread4[b >> 4];
-			dest[1] = (dest[1] >> 1) + spread4[b & 0xf];
-		   }
-	   }
+{	/* The maximum width for an EGA/VGA device is 800 pixels.... */
+	int width_bytes = (dev->width + 7) >> 3;
+	int i;
+	bits32 *dest;
+	const byte *src;
+	const byte *end;
+	byte planes[100*4];
+	/* Plane 0 is the least significant plane. */
+	/* We know we're on a little-endian machine.... */
+#define spread4(v)\
+ v+0x00000000, v+0x08000000, v+0x80000000, v+0x88000000,\
+ v+0x00080000, v+0x08080000, v+0x80080000, v+0x88080000,\
+ v+0x00800000, v+0x08800000, v+0x80800000, v+0x88800000,\
+ v+0x00880000, v+0x08880000, v+0x80880000, v+0x88880000
+	static const bits32 far_data spread8[256] =
+	{	spread4(0x0000), spread4(0x0800),
+		spread4(0x8000), spread4(0x8800),
+		spread4(0x0008), spread4(0x0808),
+		spread4(0x8008), spread4(0x8808),
+		spread4(0x0080), spread4(0x0880),
+		spread4(0x8080), spread4(0x8880),
+		spread4(0x0088), spread4(0x0888),
+		spread4(0x8088), spread4(0x8888)
+	};
+
+	if ( y < 0 || y >= dev->height || dev->width > 800 )
+	  return_error(gs_error_rangecheck);
+	/* Read 4 planes into the holding buffer. */
+	for ( i = 0; i < 4; ++i )
+	  {	set_g_read_plane(i);
+		memcpy(planes + 100 * i, mk_fb_ptr(0, y), width_bytes);
+	  }
+	/* Now assemble the final data from the planes. */
+	for ( dest = (bits32 *)data, src = planes, end = src + width_bytes;
+	      src < end; ++dest, ++src
+	    )
+	  *dest = (((((spread8[src[0]] >> 1) | spread8[src[100]]) >> 1) |
+		    spread8[src[200]]) >> 1) | spread8[src[300]];
 	if ( actual_data != 0 )
-		*actual_data = data;
+	  *actual_data = data;
 	return 0;
 }
 
