@@ -1,13 +1,24 @@
-/* Copyright (C) 1993, 1994 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1993, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gsstruct.h */
@@ -111,7 +122,8 @@ struct gs_memory_struct_type_s {
 		/* Enumerate the pointers in a structure. */
 
 #define struct_proc_enum_ptrs(proc)\
-  gs_ptr_type_t proc(P4(void /*obj_header_t*/ *ptr, uint size, uint index, void **pep))
+  gs_ptr_type_t proc(P4(void /*obj_header_t*/ *ptr, uint size, uint index,\
+    const void **pep))
 	struct_proc_enum_ptrs((*enum_ptrs));
 
 		/* Relocate all the pointers in this structure. */
@@ -122,13 +134,14 @@ struct gs_memory_struct_type_s {
 
 		/*
 		 * Finalize this structure just before freeing it.
-		 * Finalization procedures must not allocate or free
+		 * Finalization procedures must not allocate or resize
 		 * any objects in any space managed by the allocator,
-		 * and they cannot assume that any objects in such spaces
+		 * and must not assume that any objects in such spaces
 		 * referenced by this structure still exist.  However,
-		 * finalization procedures may allocate, free, and
-		 * reference objects allocated in other ways, such as
-		 * objects allocated with malloc by libraries.
+		 * finalization procedures may free such objects, and
+		 * may allocate, free, and reference objects allocated
+		 * in other ways, such as objects allocated with malloc
+		 * by libraries.
 		 */
 
 #define struct_proc_finalize(proc)\
@@ -151,6 +164,16 @@ extern_st(st_free);
 /* Define a type descriptor for byte objects. */
 extern_st(st_bytes);
 
+/* Define type descriptors for elements and arrays of const strings. */
+#define private_st_const_string()\
+  gs_private_st_composite(st_const_string, gs_const_string, "gs_const_string",\
+    const_string_enum_ptrs, const_string_reloc_ptrs)
+extern_st(st_const_string_element);
+#define public_st_const_string_element()\
+  gs_public_st_element(st_const_string_element, gs_const_string,\
+    "gs_const_string[]", const_string_elt_enum_ptrs,\
+    const_string_elt_reloc_ptrs, st_const_string)
+
 /* ================ Macros for defining structure types ================ */
 
 #define public_st public const gs_memory_struct_type_t
@@ -250,10 +273,10 @@ extern_st(st_bytes);
 
 #define gs__st_ptr(scope_st, stname, stype, sname, penum, preloc)\
   private ENUM_PTRS_BEGIN(penum) return 0;\
-    case 0: *pep = (void *)*(stype *)vptr; break;\
+    case 0: *pep = (const void *)*(stype *)vptr; return ptr_struct_type;\
   ENUM_PTRS_END\
   private RELOC_PTRS_BEGIN(preloc) ;\
-    *(stype *)vptr = gs_reloc_struct_ptr((void *)*(stype *)vptr, gcst); /* discard const */\
+    *(stype *)vptr = gs_reloc_struct_ptr((const void *)*(stype *)vptr, gcst);\
   RELOC_PTRS_END\
   gs__st_composite_only(scope_st, stname, stype, sname, penum, preloc)
 #define gs_public_st_ptr(stname, stype, sname, penum, preloc)\
@@ -285,10 +308,10 @@ RELOC_PTRS_END
  */
 #ifdef __PROTOTYPES__
 #  define ENUM_PTRS_BEGIN_PROC(proc)\
-    gs_ptr_type_t proc(void *vptr, uint size, uint index, void **pep)
+    gs_ptr_type_t proc(void *vptr, uint size, uint index, const void **pep)
 #else
 #  define ENUM_PTRS_BEGIN_PROC(proc)\
-    gs_ptr_type_t proc(vptr, size, index, pep) void *vptr; uint size; uint index; void **pep;
+    gs_ptr_type_t proc(vptr, size, index, pep) void *vptr; uint size; uint index; const void **pep;
 #endif
 #define ENUM_PTRS_BEGIN(proc)\
   ENUM_PTRS_BEGIN_PROC(proc) { switch ( index ) { default:
@@ -297,17 +320,17 @@ RELOC_PTRS_END
 #define ENUM_RETURN_PTR(typ, elt)\
   ENUM_RETURN(((typ *)vptr)->elt)
 #define ENUM_RETURN(ptr)\
-  *pep = (void *)(ptr); break		/* discard const */
+  do { *pep = (const void *)(ptr); return ptr_struct_type; } while (0)
 #define ENUM_STRING_PTR(i, typ, elt)\
   case i: ENUM_RETURN_STRING_PTR(typ, elt)
 #define ENUM_RETURN_STRING_PTR(typ, elt)\
-  *pep = (void *)&((typ *)vptr)->elt; return ptr_string_type
+  do { *pep = (const void *)&((typ *)vptr)->elt; return ptr_string_type; } while (0)
 #define ENUM_CONST_STRING_PTR(i, typ, elt)\
   case i: ENUM_RETURN_CONST_STRING_PTR(typ, elt)
 #define ENUM_RETURN_CONST_STRING_PTR(typ, elt)\
-  *pep = (void *)&((typ *)vptr)->elt; return ptr_const_string_type
+  do { *pep = (const void *)&((typ *)vptr)->elt; return ptr_const_string_type; } while (0)
 #define ENUM_PTRS_END\
-  } return ptr_struct_type; ENUM_PTRS_END_PROC }
+  } /* mustn't fall through! */ ENUM_PTRS_END_PROC }
 #define ENUM_PTRS_END_PROC /* */
 #ifdef __PROTOTYPES__
 #  define RELOC_PTRS_BEGIN(proc)\
@@ -318,7 +341,7 @@ RELOC_PTRS_END
 #endif
 #define RELOC_PTR(typ, elt)\
   ((typ *)vptr)->elt =\
-    gs_reloc_struct_ptr((void *)((typ *)vptr)->elt, gcst) /* discard const */
+    gs_reloc_struct_ptr((const void *)((const typ *)vptr)->elt, gcst)
 /* Relocate a pointer that points to a known offset within an object. */
 /* OFFSET is for byte offsets, TYPED_OFFSET is for element offsets. */
 #define RELOC_OFFSET_PTR(typ, elt, offset)\
@@ -364,6 +387,21 @@ RELOC_PTRS_END
 #define gs_private_st_ptrs1(stname, stype, sname, penum, preloc, e1)\
   gs__st_ptrs1(private_st, stname, stype, sname, penum, preloc, e1)
 
+	/* Structures with 1 pointer and 1 string. */
+
+#define gs__st_ptrs1_strings1(scope_st, stname, stype, sname, penum, preloc, e1, e2)\
+  private ENUM_PTRS_BEGIN(penum) return 0;\
+    ENUM_PTR(0,stype,e1); ENUM_STRING_PTR(1,stype,e2);\
+  ENUM_PTRS_END\
+  private RELOC_PTRS_BEGIN(preloc) ;\
+    RELOC_PTR(stype,e1); RELOC_STRING_PTR(stype,e2);\
+  RELOC_PTRS_END\
+  gs__st_composite_only(scope_st, stname, stype, sname, penum, preloc)
+#define gs_public_st_ptrs1_strings1(stname, stype, sname, penum, preloc, e1, e2)\
+  gs__st_ptrs1_strings1(public_st, stname, stype, sname, penum, preloc, e1, e2)
+#define gs_private_st_ptrs1_strings1(stname, stype, sname, penum, preloc, e1, e2)\
+  gs__st_ptrs1_strings1(private_st, stname, stype, sname, penum, preloc, e1, e2)
+
 	/* Structures with 2 pointers. */
 
 #define gs__st_ptrs2(scope_st, stname, stype, sname, penum, preloc, e1, e2)\
@@ -411,10 +449,28 @@ RELOC_PTRS_END
 #define gs_private_st_ptrs4(stname, stype, sname, penum, preloc, e1, e2, e3, e4)\
   gs__st_ptrs4(private_st, stname, stype, sname, penum, preloc, e1, e2, e3, e4)
 
+	/* Structures with 5 pointers. */
+
+#define gs__st_ptrs5(scope_st, stname, stype, sname, penum, preloc, e1, e2, e3, e4, e5)\
+  private ENUM_PTRS_BEGIN(penum) return 0;\
+    ENUM_PTR(0,stype,e1); ENUM_PTR(1,stype,e2); ENUM_PTR(2,stype,e3);\
+    ENUM_PTR(3,stype,e4); ENUM_PTR(4,stype,e5);\
+  ENUM_PTRS_END\
+  private RELOC_PTRS_BEGIN(preloc) ;\
+    RELOC_PTR(stype,e1); RELOC_PTR(stype,e2); RELOC_PTR(stype,e3);\
+    RELOC_PTR(stype,e4); RELOC_PTR(stype,e5);\
+  RELOC_PTRS_END\
+  gs__st_composite_only(scope_st, stname, stype, sname, penum, preloc)
+#define gs_public_st_ptrs5(stname, stype, sname, penum, preloc, e1, e2, e3, e4, e5)\
+  gs__st_ptrs5(public_st, stname, stype, sname, penum, preloc, e1, e2, e3, e4, e5)
+#define gs_private_st_ptrs5(stname, stype, sname, penum, preloc, e1, e2, e3, e4, e5)\
+  gs__st_ptrs5(private_st, stname, stype, sname, penum, preloc, e1, e2, e3, e4, e5)
+
 /* ---------------- Suffix subclasses ---------------- */
 
 /*
- * Boilerplate for suffix subclasses.
+ * Boilerplate for suffix subclasses.  Special subclasses constructed
+ * 'by hand' may use this also.
  */
 #define ENUM_PREFIX(supst, n)\
   return (*supst.enum_ptrs)(vptr,size,index-(n),pep)
@@ -422,16 +478,37 @@ RELOC_PTRS_END
   (*supst.reloc_ptrs)(vptr,size,gcst)
 
 	/* Suffix subclasses with no additional pointers. */
-	/* Note that unlike other subclass declarations, */
-	/* these name the superclass's enum and reloc procedures, */
-	/* rather than the superclass's descriptor. */
 
-#define gs__st_suffix_add0(scope_st, stname, stype, sname, supenum, supreloc)\
-  gs__st_composite_only(scope_st, stname, stype, sname, supenum, supreloc)
-#define gs_public_st_suffix_add0(stname, stype, sname, supenum, supreloc)\
-  gs__st_suffix_add0(public_st, stname, stype, sname, supenum, supreloc)
-#define gs_private_st_suffix_add0(stname, stype, sname, supenum, supreloc)\
-  gs__st_suffix_add0(private_st, stname, stype, sname, supenum, supreloc)
+#define gs__st_suffix_add0(scope_st, stname, stype, sname, penum, preloc, supstname)\
+  private ENUM_PTRS_BEGIN_PROC(penum) {\
+    return (*supstname.enum_ptrs)(vptr, size, index, pep);\
+  } ENUM_PTRS_END_PROC\
+  private RELOC_PTRS_BEGIN(preloc) {\
+    (*supstname.reloc_ptrs)(vptr, size, gcst);\
+  } RELOC_PTRS_END\
+  gs__st_composite_only(scope_st, stname, stype, sname, penum, preloc)
+#define gs_public_st_suffix_add0(stname, stype, sname, penum, preloc, supstname)\
+  gs__st_suffix_add0(public_st, stname, stype, sname, penum, preloc, supstname)
+#define gs_private_st_suffix_add0(stname, stype, sname, penum, preloc, supstname)\
+  gs__st_suffix_add0(private_st, stname, stype, sname, penum, preloc, supstname)
+
+	/* Suffix subclasses with no additional pointers and finalization. */
+	/* This is a hack -- subclasses should inherit finalization, */
+	/* but that would require a superclass pointer in the descriptor, */
+	/* which would perturb things too much right now. */
+
+#define gs__st_suffix_add0_final(scope_st, stname, stype, sname, penum, preloc, pfinal, supstname)\
+  private ENUM_PTRS_BEGIN_PROC(penum) {\
+    return (*supstname.enum_ptrs)(vptr, size, index, pep);\
+  } ENUM_PTRS_END_PROC\
+  private RELOC_PTRS_BEGIN(preloc) {\
+    (*supstname.reloc_ptrs)(vptr, size, gcst);\
+  } RELOC_PTRS_END\
+  gs__st_complex_only(scope_st, stname, stype, sname, 0, penum, preloc, pfinal)
+#define gs_public_st_suffix_add0_final(stname, stype, sname, penum, preloc, pfinal, supstname)\
+  gs__st_suffix_add0_final(public_st, stname, stype, sname, penum, preloc, pfinal, supstname)
+#define gs_private_st_suffix_add0_final(stname, stype, sname, penum, preloc, pfinal, supstname)\
+  gs__st_suffix_add0_final(private_st, stname, stype, sname, penum, preloc, pfinal, supstname)
 
 	/* Suffix subclasses with 1 additional pointer. */
 
@@ -448,6 +525,22 @@ RELOC_PTRS_END
 #define gs_private_st_suffix_add1(stname, stype, sname, penum, preloc, supstname, e1)\
   gs__st_suffix_add1(private_st, stname, stype, sname, penum, preloc, supstname, e1)
 
+	/* Suffix subclasses with 1 additional pointer and finalization. */
+	/* See above regarding finalization and subclasses. */
+
+#define gs__st_suffix_add1_final(scope_st, stname, stype, sname, penum, preloc, pfinal, supstname, e1)\
+  private ENUM_PTRS_BEGIN(penum) ENUM_PREFIX(supstname,1);\
+    ENUM_PTR(0,stype,e1);\
+  ENUM_PTRS_END\
+  private RELOC_PTRS_BEGIN(preloc) RELOC_PREFIX(supstname);\
+    RELOC_PTR(stype,e1);\
+  RELOC_PTRS_END\
+  gs__st_complex_only(scope_st, stname, stype, sname, 0, penum, preloc, pfinal)
+#define gs_public_st_suffix_add1_final(stname, stype, sname, penum, preloc, pfinal, supstname, e1)\
+  gs__st_suffix_add1_final(public_st, stname, stype, sname, penum, preloc, pfinal, supstname, e1)
+#define gs_private_st_suffix_add1_final(stname, stype, sname, penum, preloc, pfinal, supstname, e1)\
+  gs__st_suffix_add1_final(private_st, stname, stype, sname, penum, preloc, pfinal, supstname, e1)
+
 	/* Suffix subclasses with 2 additional pointers. */
 
 #define gs__st_suffix_add2(scope_st, stname, stype, sname, penum, preloc, supstname, e1, e2)\
@@ -463,6 +556,22 @@ RELOC_PTRS_END
 #define gs_private_st_suffix_add2(stname, stype, sname, penum, preloc, supstname, e1, e2)\
   gs__st_suffix_add2(private_st, stname, stype, sname, penum, preloc, supstname, e1, e2)
 
+	/* Suffix subclasses with 2 additional pointers and finalization. */
+	/* See above regarding finalization and subclasses. */
+
+#define gs__st_suffix_add2_final(scope_st, stname, stype, sname, penum, preloc, pfinal, supstname, e1, e2)\
+  private ENUM_PTRS_BEGIN(penum) ENUM_PREFIX(supstname,2);\
+    ENUM_PTR(0,stype,e1); ENUM_PTR(1,stype,e2);\
+  ENUM_PTRS_END\
+  private RELOC_PTRS_BEGIN(preloc) RELOC_PREFIX(supstname);\
+    RELOC_PTR(stype,e1); RELOC_PTR(stype,e2);\
+  RELOC_PTRS_END\
+  gs__st_complex_only(scope_st, stname, stype, sname, 0, penum, preloc, pfinal)
+#define gs_public_st_suffix_add2_final(stname, stype, sname, penum, preloc, pfinal, supstname, e1, e2)\
+  gs__st_suffix_add2_final(public_st, stname, stype, sname, penum, preloc, pfinal, supstname, e1, e2)
+#define gs_private_st_suffix_add2_final(stname, stype, sname, penum, preloc, pfinal, supstname, e1, e2)\
+  gs__st_suffix_add2_final(private_st, stname, stype, sname, penum, preloc, pfinal, supstname, e1, e2)
+
 	/* Suffix subclasses with 3 additional pointers. */
 
 #define gs__st_suffix_add3(scope_st, stname, stype, sname, penum, preloc, supstname, e1, e2, e3)\
@@ -478,6 +587,39 @@ RELOC_PTRS_END
 #define gs_private_st_suffix_add3(stname, stype, sname, penum, preloc, supstname, e1, e2, e3)\
   gs__st_suffix_add3(private_st, stname, stype, sname, penum, preloc, supstname, e1, e2, e3)
 
+	/* Suffix subclasses with 3 additional pointers and finalization. */
+	/* See above regarding finalization and subclasses. */
+
+#define gs__st_suffix_add3_final(scope_st, stname, stype, sname, penum, preloc, pfinal, supstname, e1, e2, e3)\
+  private ENUM_PTRS_BEGIN(penum) ENUM_PREFIX(supstname,3);\
+    ENUM_PTR(0,stype,e1); ENUM_PTR(1,stype,e2); ENUM_PTR(2,stype,e3);\
+  ENUM_PTRS_END\
+  private RELOC_PTRS_BEGIN(preloc) RELOC_PREFIX(supstname);\
+    RELOC_PTR(stype,e1); RELOC_PTR(stype,e2); RELOC_PTR(stype,e3);\
+  RELOC_PTRS_END\
+  gs__st_complex_only(scope_st, stname, stype, sname, 0, penum, preloc, pfinal)
+#define gs_public_st_suffix_add3_final(stname, stype, sname, penum, preloc, pfinal, supstname, e1, e2, e3)\
+  gs__st_suffix_add3_final(public_st, stname, stype, sname, penum, preloc, pfinal, supstname, e1, e2, e3)
+#define gs_private_st_suffix_add3_final(stname, stype, sname, penum, preloc, pfinal, supstname, e1, e2, e3)\
+  gs__st_suffix_add3_final(private_st, stname, stype, sname, penum, preloc, pfinal, supstname, e1, e2, e3)
+
+	/* Suffix subclasses with 4 additional pointers. */
+
+#define gs__st_suffix_add4(scope_st, stname, stype, sname, penum, preloc, supstname, e1, e2, e3, e4)\
+  private ENUM_PTRS_BEGIN(penum) ENUM_PREFIX(supstname,4);\
+    ENUM_PTR(0,stype,e1); ENUM_PTR(1,stype,e2); ENUM_PTR(2,stype,e3);\
+    ENUM_PTR(3,stype,e4);\
+  ENUM_PTRS_END\
+  private RELOC_PTRS_BEGIN(preloc) RELOC_PREFIX(supstname);\
+    RELOC_PTR(stype,e1); RELOC_PTR(stype,e2); RELOC_PTR(stype,e3);\
+    RELOC_PTR(stype,e4);\
+  RELOC_PTRS_END\
+  gs__st_composite_only(scope_st, stname, stype, sname, penum, preloc)
+#define gs_public_st_suffix_add4(stname, stype, sname, penum, preloc, supstname, e1, e2, e3, e4)\
+  gs__st_suffix_add4(public_st, stname, stype, sname, penum, preloc, supstname, e1, e2, e3, e4)
+#define gs_private_st_suffix_add4(stname, stype, sname, penum, preloc, supstname, e1, e2, e3, e4)\
+  gs__st_suffix_add4(private_st, stname, stype, sname, penum, preloc, supstname, e1, e2, e3, e4)
+
 /* ---------------- General subclasses ---------------- */
 
 /*
@@ -491,7 +633,7 @@ RELOC_PTRS_END
 
 	/* General subclasses with no additional pointers. */
 
-#define gs__st_ptrs_add0(scope_st, stname, stype, sname, penum, reloc, supstname, member)\
+#define gs__st_ptrs_add0(scope_st, stname, stype, sname, penum, preloc, supstname, member)\
   private ENUM_PTRS_BEGIN(penum) ENUM_SUPER(stype,supstname,member,0);\
   ENUM_PTRS_END\
   private RELOC_PTRS_BEGIN(preloc) RELOC_SUPER(stype,supstname,member);\
