@@ -1,13 +1,24 @@
-/* Copyright (C) 1992, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* scfd.c */
@@ -19,32 +30,47 @@
 #include "scf.h"
 #include "scfx.h"
 
+/* ------ CCITTFaxDecode ------ */
+
 private_st_CFD_state();
 
 #define ss ((stream_CFD_state *)st)
 
+/* Set default parameter values. */
+private void
+s_CFD_set_defaults(register stream_state *st)
+{	s_CFD_set_defaults_inline(ss);
+}
+
 /* Initialize CCITTFaxDecode filter */
 private int
 s_CFD_init(stream_state *st)
-{	int raster = ss->raster = (ss->Columns + 7) >> 3;
+{	int raster = ss->raster =
+	  round_up((ss->Columns + 7) >> 3, ss->DecodedByteAlign);
 	byte white = (ss->BlackIs1 ? 0 : 0xff);
 	s_hcd_init_inline(ss);
-	ss->lbuf = gs_alloc_bytes(st->memory, raster + 1, "CFD lbuf");
+	/* Because skip_white_pixels can look as many as 4 bytes ahead, */
+	/* we need to allow 4 extra bytes at the end of the row buffers. */
+	ss->lbuf = gs_alloc_bytes(st->memory, raster + 4, "CFD lbuf");
 	ss->lprev = 0;
 	if ( ss->lbuf == 0 )
 	  return ERRC;		/****** WRONG ******/
 	if ( ss->K != 0 )
-	{	ss->lprev = gs_alloc_bytes(st->memory, raster + 1, "CFD lprev");
-		if ( ss->lprev == 0 )
-		  return ERRC;		/****** WRONG ******/
-		/* Clear the initial reference line for 2-D encoding. */
-		memset(ss->lbuf, white, raster);
-	}
+	  { ss->lprev = gs_alloc_bytes(st->memory, raster + 4, "CFD lprev");
+	    if ( ss->lprev == 0 )
+	      return ERRC;		/****** WRONG ******/
+	    /* Clear the initial reference line for 2-D encoding. */
+	    memset(ss->lbuf, white, raster);
+	    /* Ensure that the scan of the reference line will stop. */
+	    ss->lbuf[raster] = 0xa0;
+	  }
 	ss->k_left = min(ss->K, 0);
 	ss->run_color = 0;
+	ss->damaged_rows = 0;
+	ss->skipping_damage = false;
 	ss->cbit = 0;
 	ss->uncomp_run = 0;
-	ss->rows_left = (ss->Rows <= 0 ? -1 : ss->Rows + 1);
+	ss->rows_left = (ss->Rows <= 0 || ss->EndOfBlock ? -1 : ss->Rows + 1);
 	ss->rpos = ss->wpos = raster - 1;
 	ss->eol_count = 0;
 	ss->invert = white;
@@ -127,14 +153,6 @@ s_CFD_release(stream_state *st)
 		rlen -= qbit;\
 		switch ( rlen >> 3 )\
 		{\
-		default:	/* original rlen >= 64 */\
-d:			memset(q, black_byte, rlen >> 3);\
-			q += rlen >> 3;\
-			rlen &= 7;\
-			if ( !rlen ) qbit = 0, q--;\
-			else qbit = 8 - rlen, *q ^= 0xff << qbit;\
-			makeup_action;\
-			break;\
 		case 7:		/* original rlen possibly >= 64 */\
 			if ( rlen + qbit >= 64 ) goto d;\
 			*q++ = black_byte;\
@@ -150,6 +168,14 @@ d:			memset(q, black_byte, rlen >> 3);\
 		case 0:			/* know rlen != 0 */\
 			qbit = 8 - rlen;\
 			*q ^= 0xff << qbit;\
+			break;\
+		default:	/* original rlen >= 64 */\
+d:			memset(q, black_byte, rlen >> 3);\
+			q += rlen >> 3;\
+			rlen &= 7;\
+			if ( !rlen ) qbit = 0, q--;\
+			else qbit = 8 - rlen, *q ^= 0xff << qbit;\
+			makeup_action;\
 		}\
 	}\
 	else\
@@ -169,11 +195,61 @@ s_CFD_process(stream_state *st, stream_cursor_read *pr,
 	int k_left = ss->k_left;
 	int rows_left = ss->rows_left;
 	int status = 0;
-top:	/* Check for a completed scan line. */
-	/* We must make a special check for a 2-D line that ends with */
-	/* a horizontal code with a 0-length second run! */
+#ifdef DEBUG
+	const byte *rstart = pr->ptr;
+	const byte *wstart = pw->ptr;
+#endif
+
+top:
+#ifdef DEBUG
+	{ hcd_declare_state;
+	  hcd_load_state();
+	  if_debug8('w', "\
+[w]CFD_process top: eol_count=%d, k_left=%d, rows_left=%d\n\
+    bits=0x%lx, bits_left=%d, read %u, wrote %u%s\n",
+		    eol_count, k_left, rows_left,
+		    (ulong)bits, bits_left,
+		    (uint)(p - rstart), (uint)(pw->ptr - wstart),
+		    (ss->skipping_damage ? ", skipping damage" : ""));
+	}
+#endif
+	if ( ss->skipping_damage )
+	  {	/* Skip until we reach an EOL. */
+		hcd_declare_state;
+		int skip;
+		status = 0;
+		do
+		  { switch ( (skip = cf_decode_eol(ss, pr)) )
+		      {
+		      default:		/* not EOL */
+			hcd_load_state();
+			skip_bits(-skip);
+			hcd_store_state();
+			continue;
+		      case 0:		/* need more input */
+			goto out;
+		      case 1:		/* EOL */
+			{ /* Back up over the EOL. */
+			  hcd_load_state();
+			  bits_left += run_eol_code_length;
+			  hcd_store_state();
+			}
+			ss->skipping_damage = false;
+		      }
+		  }
+		while ( ss->skipping_damage );
+		ss->damaged_rows++;
+	  }
+	/*
+	 * Check for a completed input scan line.  This isn't quite as
+	 * simple as it seems, because we could have run out of input data
+	 * between a makeup code and a 0-length termination code, or in a
+	 * 2-D line before a final horizontal code with a 0-length second
+	 * run.  There's probably a way to think about this situation that
+	 * doesn't require a special check, but I haven't found it yet.
+	 */
 	if ( ss->wpos == wstop && ss->cbit <= (-ss->Columns & 7) &&
-	     !(k_left != 0 && ss->run_color != 0)
+	     (k_left == 0 ? !(ss->run_color & ~1) : ss->run_color == 0)
 	   )
 	{	/* Check for completed data to be copied to the client. */
 		/* (We could avoid the extra copy step for 1-D, but */
@@ -204,8 +280,13 @@ top:	/* Check for a completed scan line. */
 		ss->invert = (ss->BlackIs1 ? 0 : 0xff);
 		memset(ss->lbuf, ss->invert, wstop + 1);
 		ss->run_color = 0;
-		if ( ss->EncodedByteAlign )
-			ss->bits_left &= ~7;
+		/*
+		 * If EndOfLine is true, we want to include the byte padding
+		 * in the string of initial zeros in the EOL.  If EndOfLine
+		 * is false, we aren't sure what we should do....
+		 */
+		if ( ss->EncodedByteAlign & !ss->EndOfLine )
+		  ss->bits_left &= ~7;
 	}
 	/* If we're between scan lines, scan for EOLs. */
 	if ( ss->wpos < 0 )
@@ -240,13 +321,13 @@ top:	/* Check for a completed scan line. */
 			if ( ss->EndOfLine )
 			{	/* EOL is required, but none is present. */
 				status = ERRC;
-				goto out;
+				goto check;
 			}
 		case 1:
 			break;
 		default:
 			status = ERRC;
-			goto out;
+			goto check;
 		}
 	}
 	/* Now decode actual data. */
@@ -264,8 +345,27 @@ top:	/* Check for a completed scan line. */
 	}
 	if_debug3('w', "[w]CFD status = %d, wpos = %d, cbit = %d\n",
 		  status, ss->wpos, ss->cbit);
-	if ( status == 1 )
+check:	switch ( status )
+	  {
+	  case 1:			/* output full */
 		goto top;
+	  case ERRC:
+		/* Check for special handling of damaged rows. */
+		if ( ss->damaged_rows >= ss->DamagedRowsBeforeError ||
+		     !(ss->EndOfLine && ss->K >= 0)
+		   )
+		  break;
+		/* Substitute undamaged data if appropriate. */
+		/****** NOT IMPLEMENTED YET ******/
+		{ ss->wpos = wstop;
+		  ss->cbit = -ss->Columns & 7;
+		  ss->run_color = 0;
+		}
+		ss->skipping_damage = true;
+		goto top;
+	  default:
+		ss->damaged_rows = 0;	/* finished a good row */
+	  }
 out:	ss->k_left = k_left;
 	ss->rows_left = rows_left;
 	ss->eol_count = eol_count;
@@ -274,9 +374,16 @@ out:	ss->k_left = k_left;
 
 #undef ss
 
-/* Decode a leading EOL, if any. */
-/* If an EOL is present, skip over it and return 1; */
-/* if no EOL is present, return -1; if more input is needed, return 0. */
+/*
+ * Decode a leading EOL, if any.
+ * If an EOL is present, skip over it and return 1;
+ * if no EOL is present, read no input and return -N, where N is the
+ * number of initial bits that can be skipped in the search for an EOL;
+ * if more input is needed, return 0.
+ * Note that if we detected an EOL, we know that we can back up over it;
+ * if we detected an N-bit non-EOL, we know that at least N bits of data
+ * are available in the buffer.
+ */
 private int
 cf_decode_eol(stream_CFD_state *ss, stream_cursor_read *pr)
 {	hcd_declare_state;
@@ -286,7 +393,7 @@ cf_decode_eol(stream_CFD_state *ss, stream_cursor_read *pr)
 	for ( zeros = 0; zeros < run_eol_code_length - 1; zeros++ )
 	{	ensure_bits(1, out);
 		if ( peek_bits(1) )
-		  return -1;
+		  return -(zeros + 1);
 		skip_bits(1);
 	}
 	/* We definitely have an EOL.  Skip further zero bits. */
@@ -300,10 +407,12 @@ cf_decode_eol(stream_CFD_state *ss, stream_cursor_read *pr)
 	skip_bits(1);
 	hcd_store_state();
 	return 1;
-back:	/* We ran out of data while skipping zeros. */
-	/* We know we are at a byte boundary, and have just skipped */
-	/* at least run_eol_code_length - 1 zeros. */
-	/* However, bits_left may be 1 if look_ahead == 2. */
+back:	/*
+	 * We ran out of data while skipping zeros.
+	 * We know we are at a byte boundary, and have just skipped
+	 * at least run_eol_code_length - 1 zeros.  However,
+	 * bits_left may be 1 if look_ahead == 2.
+	 */
 	bits &= (1 << bits_left) - 1;
 	bits_left += run_eol_code_length - 1;
 	hcd_store_state();
@@ -317,16 +426,20 @@ cf_decode_1d(stream_CFD_state *ss, stream_cursor_read *pr)
 	byte black_byte = (ss->BlackIs1 ? 0xff : 0);
 	int end_bit = -ss->Columns & 7;
 	byte *stop = ss->lbuf - 1 + ss->raster;
+	int run_color = ss->run_color;
 	int status;
 	int bcnt;
+
 	cfd_load_state();
-	if ( ss->run_color )
-		goto db;
+	if_debug1('w', "[w1]entry run_color = %d\n", ss->run_color);
+	if ( ss->run_color > 0 )
+	  goto db;
 	else
-		goto dw;
+	  goto dw;
 #define q_at_stop() (q >= stop && (qbit <= end_bit || q > stop))
-top:	if ( q_at_stop() )
-		goto done;
+top:	run_color = 0;
+	if ( q_at_stop() )
+	  goto done;
 dw:	/* Decode a white run. */
 	get_run(cf_white_decode, cfd_white_initial_bits, bcnt, "[w1]white", out0);
 	if ( bcnt < 0 )			/* exceptional situation */
@@ -336,7 +449,7 @@ dw:	/* Decode a white run. */
 			cfd_store_state();
 			bcnt = cf_decode_uncompressed(ss, pr);
 			if ( bcnt < 0 )
-				return bcnt;
+			  return bcnt;
 			cfd_load_state();
 			if ( bcnt ) goto db;
 			else goto dw;
@@ -347,9 +460,12 @@ dw:	/* Decode a white run. */
 			goto out;
 		}
 	}
-	skip_data(bcnt, dw);
+	skip_data(bcnt, dwx);
 	if ( q_at_stop() )
+	  {	run_color = 0;		/* not inside a run */
 		goto done;
+	  }
+	run_color = 1;
 db:	/* Decode a black run. */
 	get_run(cf_black_decode, cfd_black_initial_bits, bcnt, "[w1]black", out1);
 	if ( bcnt < 0 )
@@ -359,18 +475,28 @@ db:	/* Decode a black run. */
 		goto out;
 	}
 	/* Invert bits designated by black run. */
-	invert_data(bcnt, black_byte, goto db, idb);
+	invert_data(bcnt, black_byte, goto dbx, idb);
 	goto top;
+dwx:	/* If we run out of data after a makeup code, */
+	/* note that we are still processing a white run. */
+	run_color = -1;
+	goto dw;
+dbx:	/* If we run out of data after a makeup code, */
+	/* note that we are still processing a black run. */
+	run_color = 2;
+	goto db;
 done:	if ( q > stop || qbit < end_bit )
-		status = ERRC;
+	  status = ERRC;
 	else
-		status = 1;
+	  status = 1;
 out:	cfd_store_state();
+	ss->run_color = run_color;
+	if_debug1('w', "[w1]exit run_color = %d\n", run_color);
 	return status;
-out0:	ss->run_color = 0;
+out0:	/* We already set run_color to 0 or -1. */
 	status = 0;
 	goto out;
-out1:	ss->run_color = 1;
+out1:	/* We already set run_color to 1 or 2. */
 	status = 0;
 	goto out;
 }
@@ -398,8 +524,10 @@ cf_decode_2d(stream_CFD_state *ss, stream_cursor_read *pr)
 	if_debug1('W', "[w2]raster=%d\n", raster);
 	switch ( ss->run_color )
 	{
+	case -2: ss->run_color = 0; goto hww;
 	case -1: ss->run_color = 0; goto hbw;
 	case 1: ss->run_color = 0; goto hwb;
+	case 2: ss->run_color = 0; goto hbb;
 	/*case 0:*/
 	}
 top:	if ( count <= end_count )
@@ -438,27 +566,9 @@ top:	if ( count <= end_count )
 	case 1:					/* horizontal */
 		skip_bits(3);
 		if ( invert == invert_white )
-		{	/* White, then black. */
-hww:			get_run(cf_white_decode, cfd_white_initial_bits,
-				rlen, " white", hback);
-			if ( (count -= rlen) < end_count )
-			{	status = ERRC;
-				goto out;
-			}
-			skip_data(rlen, hww);
-			goto hwb;
-		}
+		  goto hww;
 		else
-		{	/* Black, then white. */
-hbb:			get_run(cf_black_decode, cfd_black_initial_bits,
-				rlen, " black", hback);
-			if ( (count -= rlen) < end_count )
-			{	status = ERRC;
-				goto out;
-			}
-			invert_data(rlen, black_byte, goto hbb, ihbb);
-			goto hbw;
-		}
+		  goto hbb;
 	case 0:				/* everything else */
 		get_run(cf_2d_decode, cfd_2d_initial_bits, rlen,
 			"[w2]", out0);
@@ -560,44 +670,54 @@ hbb:			get_run(cf_black_decode, cfd_black_initial_bits,
 out:	cfd_store_state();
 	ss->invert = invert;
 	return status;
-		/* Handle input exhausted after recognizing a horizontal */
-		/* mode, but before parsing the black or white run. */
-		/* The mode flag is 3 bits; the longest black/white */
-		/* run code is 13 bits, so bits_left <= 12, and we can */
-		/* back up by just incrementing bits_left. */
-		/* To make this work even with makeup codes, */
-		/* we force the previous 3 bits to 001 (horizontal mode). */
-hback:	bits = (bits & ((1 << bits_left) - 1)) + (1 << bits_left);
-	bits_left += 3;
 out0:	status = 0;
 	goto out;
-		/* Handle input exhausted when parsing the second run */
-		/* of a horizontal sequence.  We have to resume parsing */
-		/* at hwb or hbw depending on the run color. */
-outbw:	ss->run_color = -1;
-	goto out0;
-outwb:	ss->run_color = 1;
-	goto out0;
+		/*
+		 * We handle horizontal decoding here, so that we can
+		 * branch back into it if we run out of input data.
+		 */
+		/* White, then black. */
+hww:	get_run(cf_white_decode, cfd_white_initial_bits, rlen,
+		" white", outww);
+	if ( (count -= rlen) < end_count )
+	  { status = ERRC;
+	    goto out;
+	  }
+	skip_data(rlen, hww);
 		/* Handle the second half of a white-black horizontal code. */
-		/* We put this here so we can branch to it when resuming. */
 hwb:	get_run(cf_black_decode, cfd_black_initial_bits, rlen,
 		" black", outwb);
 	if ( (count -= rlen) < end_count )
-	{	status = ERRC;
-		goto out;
-	}
+	  { status = ERRC;
+	    goto out;
+	  }
 	invert_data(rlen, black_byte, goto hwb, ihwb);
 	goto top;
+outww:	ss->run_color = -2;
+	goto out0;
+outwb:	ss->run_color = 1;
+	goto out0;
+		/* Black, then white. */
+hbb:	get_run(cf_black_decode, cfd_black_initial_bits, rlen,
+		" black", outbb);
+	if ( (count -= rlen) < end_count )
+	  { status = ERRC;
+	    goto out;
+	  }
+	invert_data(rlen, black_byte, goto hbb, ihbb);
 		/* Handle the second half of a black-white horizontal code. */
-		/* We put this here so we can branch to it when resuming. */
 hbw:	get_run(cf_white_decode, cfd_white_initial_bits, rlen,
 		" white", outbw);
 	if ( (count -= rlen) < end_count )
-	{	status = ERRC;
-		goto out;
-	}
+	  { status = ERRC;
+	    goto out;
+	  }
 	skip_data(rlen, hbw);
 	goto top;
+outbb:	ss->run_color = 2;
+	goto out0;
+outbw:	ss->run_color = -1;
+	goto out0;
 }
 
 #if 1	/****************/
@@ -661,5 +781,6 @@ out:		/******* WRONG ******/
 
 /* Stream template */
 const stream_template s_CFD_template =
-{	&st_CFD_state, s_CFD_init, s_CFD_process, 1, 1, s_CFD_release
+{	&st_CFD_state, s_CFD_init, s_CFD_process, 1, 1, s_CFD_release,
+	s_CFD_set_defaults
 };
