@@ -1,13 +1,24 @@
-/* Copyright (C) 1994 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1994, 1995 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* spdiffx.h */
@@ -32,5 +43,7 @@ typedef struct stream_PDiff_state_s {
 #define private_st_PDiff_state()	/* in spdiff.c */\
   gs_private_st_simple(st_PDiff_state, stream_PDiff_state,\
     "PixelDifferenceEncode/Decode state")
+#define s_PDiff_set_defaults_inline(ss)\
+  ((ss)->Colors = 1, (ss)->BitsPerComponent = 8, (ss)->Columns = 1)
 extern const stream_template s_PDiffD_template;
 extern const stream_template s_PDiffE_template;
