@@ -1,13 +1,24 @@
-/* Copyright (C) 1992, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* zht2.c */
@@ -24,12 +35,9 @@
 #include "idparam.h"
 #include "igstate.h"
 #include "icolor.h"
+#include "iht.h"
 #include "store.h"
 
-/* Imported from zht.c */
-int zscreen_enum_init(P6(os_ptr, const gx_ht_order *, gs_screen_halftone *,
-			 ref *, int, int (*)(P1(os_ptr))));
-
 /* Forward references */
 private int dict_spot_params(P4(const ref *, gs_spot_halftone *,
   ref *, ref *));
@@ -37,30 +45,6 @@ private int dict_spot_results(P2(ref *, const gs_spot_halftone *));
 private int dict_threshold_params(P3(const ref *,
   gs_threshold_halftone *, ref *));
 
-/* Structure types */
-private_st_halftone_component();
-gs_private_st_element(st_halftone_component_element, gs_halftone_component,
-  "gs_halftone_component[]", ht_comp_elt_enum_ptrs, ht_comp_elt_reloc_ptrs,
-  st_halftone_component);
-
-/* GC procedures */
-
-#define hptr ((gs_halftone_component *)vptr)
-
-private ENUM_PTRS_BEGIN(halftone_component_enum_ptrs) return 0;
-	case 0:
-		if ( hptr->type != ht_type_threshold )
-		  return 0;
-		ENUM_RETURN_CONST_STRING_PTR(gs_halftone_component, params.threshold.thresholds);
-ENUM_PTRS_END
-
-private RELOC_PTRS_BEGIN(halftone_component_reloc_ptrs) {
-	if ( hptr->type == ht_type_threshold )
-	  RELOC_CONST_STRING_PTR(gs_halftone_component, params.threshold.thresholds);
-} RELOC_PTRS_END
-
-#undef hptr
-
 /* <dict> <dict5> .sethalftone5 - */
 float spot_dummy(P2(floatp, floatp));		/* in zht1.c */
 int spot_sample_finish(P1(os_ptr));		/* in zht.c */
@@ -79,10 +63,13 @@ zsethalftone5(register os_ptr op)
 	 { gs_ht_separation_name_strings };
 	ref sprocs[countof(color_names)];
 	ref tprocs[countof(color_names)];
+	gs_memory_t *mem;
 	int npop = 2;
 
 	check_type(*op, t_dictionary);
 	check_dict_read(*op);
+	check_type(op[-1], t_dictionary);
+	check_dict_read(op[-1]);
 	count = 0;
 	for ( i = 0; i < countof(color_names); i++ )
 	{	ref *pvalue;
@@ -91,15 +78,17 @@ zsethalftone5(register os_ptr op)
 		else if ( i == gs_ht_separation_Default )
 			return_error(e_typecheck);
 	}
+	mem = (gs_memory_t *)idmemory->spaces.indexed[r_space_index(op - 1)];
 	check_estack(5);		/* for sampling Type 1 screens */
 	refset_null(sprocs, countof(sprocs));
 	refset_null(tprocs, countof(tprocs));
-	pht = ialloc_struct(gs_halftone, &st_halftone, ".sethalftone5");
-	phtc = ialloc_struct_array(count, gs_halftone_component,
-				   &st_halftone_component_element,
-				   ".sethalftone5");
-	pdht = ialloc_struct(gx_device_halftone, &st_device_halftone,
-			     ".sethalftone5");
+	rc_alloc_struct_0(pht, gs_halftone, &st_halftone,
+			  imemory, pht = 0, ".sethalftone5");
+	phtc = gs_alloc_struct_array(mem, count, gs_halftone_component,
+				     &st_ht_component_element,
+				     ".sethalftone5");
+	rc_alloc_struct_0(pdht, gx_device_halftone, &st_device_halftone,
+			  imemory, pdht = 0, ".sethalftone5");
 	if ( pht == 0 || phtc == 0 || pdht == 0 )
 	  code = gs_note_error(e_VMerror);
 	else
@@ -136,12 +125,21 @@ zsethalftone5(register os_ptr op)
 			}
 			if ( code < 0 )
 				break;
-			pc->cname = i;
+			pc->cname = (gs_ht_separation_name)i;
 			pc++, j++;
 		}
 	}
 	if ( code >= 0 )
-	  {	pht->type = ht_type_multiple;
+	  {	/* We think that Type 2 and Type 4 halftones, like */
+		/* screens set by setcolorscreen, adapt automatically to */
+		/* the device color space, so we need to mark them */
+		/* with a different internal halftone type. */
+		int type = 0;
+
+		dict_int_param(op - 1, "HalftoneType", 1, 5, 0, &type);
+		pht->type =
+		  (type == 2 || type == 4 ? ht_type_multiple_colorscreen :
+		   ht_type_multiple);
 		pht->params.multiple.components = phtc;
 		pht->params.multiple.num_comp = count;
 		code = gs_sethalftone_prepare(igs, pht, pdht);
@@ -182,7 +180,7 @@ zsethalftone5(register os_ptr op)
 			  case ht_type_spot:
 				code = zscreen_enum_init(op, porder,
 					&phtc[j].params.spot.screen,
-					&sprocs[j], 0, 0);
+					&sprocs[j], 0, 0, mem);
 				if ( code < 0 )
 					break;
 				/* falls through */
@@ -215,9 +213,9 @@ zsethalftone5(register os_ptr op)
 		  }
 	  }
 	if ( code < 0 )
-	  {	ifree_object(pdht, ".sethalftone5");
-		ifree_object(phtc, ".sethalftone5");
-		ifree_object(pht, ".sethalftone5");
+	  {	gs_free_object(mem, pdht, ".sethalftone5");
+		gs_free_object(mem, phtc, ".sethalftone5");
+		gs_free_object(mem, pht, ".sethalftone5");
 		return code;
 	  }
 	pop(npop);
@@ -241,10 +239,13 @@ sethalftone_finish(os_ptr op)
 /* Clean up after installing the halftone. */
 private int
 sethalftone_cleanup(os_ptr op)
-{	ifree_object(esp[4].value.pstruct,
-		     "sethalftone_cleanup(device halftone)");
-	ifree_object(esp[3].value.pstruct,
-		     "sethalftone_cleanup(halftone)");
+{	gx_device_halftone *pdht = r_ptr(&esp[4], gx_device_halftone);
+	gs_halftone *pht = r_ptr(&esp[3], gs_halftone);
+
+	gs_free_object(pdht->rc.memory, pdht,
+		       "sethalftone_cleanup(device halftone)");
+	gs_free_object(pht->rc.memory, pht,
+		       "sethalftone_cleanup(halftone)");
 	return 0;
 }
 
@@ -271,7 +272,8 @@ dict_spot_params(const ref *pdict, gs_spot_halftone *psp,
 	     (code = dict_float_param(pdict, "Angle", 0.0,
 				      &psp->screen.angle)) != 0 ||
 	     (code = dict_proc_param(pdict, "SpotFunction", psproc, false)) != 0 ||
-	     (code = dict_bool_param(pdict, "AccurateScreens", false,
+	     (code = dict_bool_param(pdict, "AccurateScreens",
+				     gs_currentaccuratescreens(),
 				     &psp->accurate_screens)) < 0 ||
 	     (code = dict_proc_param(pdict, "TransferFunction", ptproc, false)) < 0
 	   )
